-------------------------------------------------------------------------------
--! @file   AEZ_E.vhd
--! @brief  AEZ_E module
--!
--!         dcalc_mode
--!             determines the precalculation operation
--!             and the subsequent operation of the start signal
--!
--!                            | Pre-calc           | Operation
--!             -----|---------|------|-----|-------|-------|------
--!             mode | (j, i)  | init | i/j | Alpha | delta | Key
--!                  |         | val  | RAM |   ?   | final |
--!             -----|---------|------|-----|-------|-------|------
--!               0  | ( 0, x) | 0    |  I  |  No   |  No   | K1 (0)
--!               1  | ( 1, x) | 0    |  I  |  Yes  |  No   | K1 (0)
--!               2  | ( 2, x) | 0    |  I  |  Yes  |  No   | K2 (1)
--!               3  | ( 3, 1) | L    |  J  |  Yes  |  Yes  | K1 (0)
--!               4  | ( 4, 0) | 2L   |  J  |  No   |  Yes  | K1 (0)
--!               5  | ( 5, 0) | 4L   |  J  |  No   |  Yes  | K1 (0)
--!               6  | ( 5, X) | 4l   |  J  |  Yes  |  Yes  | K1 (0)
--!               7  | (-1, x) | 0    |  J  |  No   |  No   | K10(2)
--!
--!             A standard operation would look like this
--!
--!             dcalc_1 | dcalc_2 | dcalc_3 | ...
--!                     | Op_1    | Op_2    | ...
--!
--!         dcalc_init  = i mod 8
--!
--!         block_number should be synchronized with pre-calc operation
--!
--!         Note: Alpha = 2^(3+floor(block_number/8))
--!               block_number = i-1
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use work.AES_pkg.all;

entity AEZ_E is
    port    (
        clk                 : in  std_logic;
        rst                 : in  std_logic;
        --! Input
        key                 : in  std_logic_vector( 32          -1 downto 0);
        xx                  : in  std_logic_vector(128          -1 downto 0);
        --! Pre-Processor Control
        key_valid           : in  std_logic;
        key_ready           : out std_logic;
        --! Control
        start               : in  std_logic;
        clr_state           : in  std_logic;
        init_key            : in  std_logic;
        dcalc_mode          : in  std_logic_vector(  3          -1 downto 0);
        dcalc_init          : in  std_logic_vector(  3          -1 downto 0);
        block_number        : in  std_logic_vector(  8          -1 downto 0);
        --! Status
        init_key_done       : out std_logic;
        almost_done         : out std_logic;
        done                : out std_logic;
        ready               : out std_logic;
        --! Output
        yy                  : out std_logic_vector(128          -1 downto 0)
    );
end entity AEZ_E;

architecture structure of AEZ_E is
    constant E_RND                  : integer := 4;
    constant E_RND_BIG              : integer := 10;

    --! Control signals
    --!     Status
    signal is_alpha                 : std_logic;
    signal is_final                 : std_logic;
    signal is_i_done                : std_logic;
    signal mode_dcalc_r             : std_logic_vector(  3          -1 downto 0);
    signal compute_round_r          : std_logic_vector(  4          -1 downto 0);
    signal round                    : std_logic_vector(  4          -1 downto 0);
    --!     from controller
    signal almost_done_s            : std_logic;
    signal done_s                   : std_logic;
    signal init_key_done_s          : std_logic;
    signal en_key                   : std_logic;
    signal en_key_init              : std_logic;
    signal en_dcalc_start           : std_logic;
    signal en_dcalc_stage           : std_logic_vector(  3          -1 downto 0);
    signal en_dcalc                 : std_logic;
    signal en_state                 : std_logic;
    signal en_rami                  : std_logic;
    signal en_ramj                  : std_logic;
    signal en_round                 : std_logic;
    signal clr_round                : std_logic;
    signal sel_final                : std_logic;
    signal sel_key_init             : std_logic_vector(  2          -1 downto 0);
    signal sel_round_type           : std_logic_vector(  2          -1 downto 0);
    signal done_ram_init            : std_logic;
    --!     derived
    signal sel_j                    : std_logic;
    signal sel_dcalc                : std_logic_vector(  3          -1 downto 0);
    signal sel_roundkey             : std_logic_vector(  2          -1 downto 0);
    signal rom_roundkey_addr        : std_logic_vector(  6          -1 downto 0);
    signal ram_addr                 : std_logic_vector(  4          -1 downto 0);
    --!     State
    type    state_type is (S_IDLE, S_LD_KEY, S_INIT_I, S_SETUP_IJ, S_COMPUTE);
    signal state                    : state_type;
    signal nstate                   : state_type;

    --! Data signals
    --!     signals
    signal alpha_r                  : std_logic_vector(  4          -1 downto 0);
    signal key_r                    : std_logic_vector(384          -1 downto 0);
    signal key_ii                   : std_logic_vector(128          -1 downto 0);
    signal key_jj                   : std_logic_vector(128          -1 downto 0);
    signal key_ll                   : std_logic_vector(128          -1 downto 0);
    signal double_key_ll            : std_logic_vector(128          -1 downto 0);
    signal double_double_key_ll     : std_logic_vector(128          -1 downto 0);
    signal key_init_r               : std_logic_vector(128          -1 downto 0);
    signal selected_key             : std_logic_vector(128          -1 downto 0);
    signal double_key               : std_logic_vector(128          -1 downto 0);
    signal delta_enc_in             : std_logic_vector(128          -1 downto 0);
    signal oo_aes                   : std_logic_vector(128          -1 downto 0);
    signal ram_data                 : std_logic_vector(128          -1 downto 0);
    signal roundkey                 : std_logic_vector(128          -1 downto 0);
    signal xor_final                : std_logic_vector(128          -1 downto 0);
    --!     registers
    signal delta_enc_r              : std_logic_vector(128          -1 downto 0);
    signal delta_enc_r2             : std_logic_vector(128          -1 downto 0);
    signal state_r                  : std_logic_vector(128          -1 downto 0);
    --!     mapped signals
    signal state_r_mapped           : t_AES_state;
    signal roundkey_mapped          : t_AES_state;
    signal oo_aes_mapped            : t_AES_state;


    --! RAM [0..7] = I, [8..15] = J
    type ram_type is array(0 to 15) of std_logic_vector(128         -1 downto 0);
    signal rami : ram_type;
    signal ramj : ram_type;

    --! Rom
    type rom_type is array(0 to 63) of std_logic_vector(  2         -1 downto 0);
    constant rom_roundkey : rom_type :=
        ("01", "00", "10", "11", "00", "00", "00", "00", --![ 0- 3] = AEZ4_k1
         "00", "00", "00", "00", "00", "00", "00", "00", --![ 4-15] = XX (00)
         "10", "00", "01", "10", "00", "00", "00", "00", --![16-19] = AEZ4_k2
         "00", "00", "00", "00", "00", "00", "00", "00", --![20-31] = XX (00)
         "00", "01", "10", "00", "01", "10", "00", "01", --![32-42] = AEZ10
         "10", "00", "00", "00", "00", "00", "00", "00", --![43-63] = XX (00)
         "00", "00", "00", "00", "00", "00", "00", "00",
         "00", "00", "00", "00", "00", "00", "00", "00");

    --! Function
    function double_block(xx: std_logic_vector(128-1 downto 0))
    return std_logic_vector is
        variable shifted_xx : std_logic_vector(128-1 downto 0);
        variable msb        : std_logic;
        variable output     : std_logic_vector(128-1 downto 0);
    begin
        shifted_xx := xx(126 downto 0) & '0';
        msb        := xx(127);
        output :=   (shifted_xx(127 downto 8)) &
                    (shifted_xx(7) xor msb)    &
                    (shifted_xx(6 downto 3))   &
                    (shifted_xx(2) xor msb)    &
                    (shifted_xx(1) xor msb)    &
                    (shifted_xx(0) xor msb);
        return output;
    end function double_block;
begin
    --! ==============
    --! Datapath
    --! ==============
    --! Internal signals
    key_ii              <= key_r(384-1 downto 256);
    key_jj              <= key_r(256-1 downto 128);
    key_ll              <= key_r(128-1 downto   0);

    --! Registers
    pReg: process(clk)
    begin
        if rising_edge(clk) then
            --! Key register
            if (en_key = '1') then
                key_r <= key_r(351 downto 0) & key;
            end if;
            --! Key doubling initialization register
            if (en_key_init = '1') then
                key_init_r <= selected_key;
            end if;
            --! Delta pre-computation register
            if (en_dcalc = '1') then
                delta_enc_r <= delta_enc_in;
            end if;
            --! Delta delay register with reset
            if (start = '1') then
                delta_enc_r2 <= delta_enc_r;
            end if;
            --! State register
            if (clr_state = '1') then
                state_r <= (others => '0');
            elsif (start = '1') then
                state_r <= delta_enc_r xor xx;
            elsif (en_state = '1') then
                state_r <= oo_aes xor xor_final;
            end if;
            --! Key doubling RAM
            if (en_rami = '1') then
                rami(to_integer(unsigned(ram_addr))) <= key_init_r;
            end if;
            if (en_ramj = '1') then
                ramj(to_integer(unsigned(ram_addr))) <= key_init_r;
            end if;
        end if;
    end process;

    --! ================ Previous block operation
    --! Key doubling pre-computation
    with sel_key_init select
    selected_key <= key_ii   when "10",
                    key_jj   when "11",
                    double_key when others;
    double_key   <= double_block(key_init_r);

    ram_data <= rami(to_integer(unsigned(ram_addr))) when sel_j = '0' else
                ramj(to_integer(unsigned(ram_addr)));

    --! Delta 1 input selection
    with sel_dcalc select
    delta_enc_in <= delta_enc_r xor ram_data   when "000",
                    key_ll                     when "001",
                    double_key_ll              when "010",
                    double_double_key_ll       when "011",
                    (others => '0')            when others;
    double_key_ll        <= double_block(key_ll);
    double_double_key_ll <= double_block(double_key_ll);

    --! ================ Current block operation
    --! I and J RAM addr selection
    ram_addr <= alpha_r when (state = S_COMPUTE and round = "0011") else round;

    --! Main
    uAes: entity work.AES_Round(basic)
    port map (din       => state_r_mapped,
              rkey      => roundkey_mapped,
              dout_fdb  => oo_aes_mapped,
              dout      => open
              );


    u_map_din: entity work.AES_map(structure)
    port map (ii => state_r,
              oo => state_r_mapped);
    u_map_key: entity work.AES_map(structure)
    port map (ii => roundkey,
              oo => roundkey_mapped);
    u_invmap: entity work.AES_invmap(structure)
    port map (ii => oo_aes_mapped,
              oo => oo_aes);

    with sel_roundkey(1 downto 0) select
    roundkey <= key_ii          when "00",
                key_jj          when "01",
                key_ll          when "10",
                (others => '0') when others;

    xor_final         <= delta_enc_r2 when sel_final = '1' else (others => '0');
    rom_roundkey_addr <= sel_round_type & round;
    sel_roundkey      <= rom_roundkey(to_integer(unsigned(rom_roundkey_addr)));

    --! Output
    yy     <= state_r;


    --! ==============
    --! Controller
    --! ==============
    pCtrlReg: process(clk)
    begin
        if rising_edge(clk) then
            if (rst = '1') then
                state <= S_IDLE;
            else
                state <= nstate;
            end if;

            if (clr_round = '1') then
                round <= (others => '0');
            elsif (en_round = '1') then
                round <= std_logic_vector(unsigned(round) + natural(1));
            end if;

            if (rst = '1') then
                en_dcalc_stage <= (others => '0');
            elsif (start = '1') then
                en_dcalc_stage <= dcalc_init;
            elsif (en_round = '1') then
                en_dcalc_stage <= '0' & en_dcalc_stage(2 downto 1);
            end if;

            if (rst = '1') then
                mode_dcalc_r <= "001";
            elsif (start = '1') then
                mode_dcalc_r <= dcalc_mode;
            end if;

            if (rst = '1') then
                is_i_done <= '0';
            elsif (done_ram_init = '1') then
                is_i_done <= not is_i_done;
            end if;

            --! Pre-calculation related status registers
            if (start = '1') then
                if (unsigned(dcalc_mode) > natural(2))
                then
                    sel_j <= '1';
                else
                    sel_j <= '0';
                end if;

                if (unsigned(dcalc_mode) = natural(1)
                    or unsigned(dcalc_mode) = natural(2)
                    or unsigned(dcalc_mode) = natural(3)
                    or unsigned(dcalc_mode) = natural(6))
                then
                    is_alpha <= '1';
                else
                    is_alpha <= '0';
                end if;

                alpha_r  <= std_logic_vector(unsigned(block_number(6 downto 3)) + natural(3));
            end if;

            --! Operation related status
            if (start = '1') then
                if (unsigned(mode_dcalc_r) = natural(7)) then
                    compute_round_r <= std_logic_vector(to_unsigned(E_RND_BIG-1,
                                            compute_round_r'length));
                else
                    compute_round_r <= std_logic_vector(to_unsigned(E_RND-1,
                                            compute_round_r'length));
                end if;

                if (unsigned(mode_dcalc_r) = natural(7)) then
                    sel_round_type <= "10";
                elsif (unsigned(mode_dcalc_r) = natural(2)) then
                    sel_round_type <= "01";
                else
                    sel_round_type <= (others => '0');
                end if;

                if (unsigned(mode_dcalc_r) > natural(2)
                    and unsigned(mode_dcalc_r) < natural(7))
                then
                    is_final <= '1';
                else
                    is_final <= '0';
                end if;
            end if;


            --! Output status
            if (rst = '1') then
                ready <= '1';
            elsif (start = '1') then
                ready <= '0';
            elsif (done_s = '1') then
                ready <= '1';
            end if;
            done            <= done_s;
            almost_done     <= almost_done_s;
            init_key_done   <= init_key_done_s;
        end if;
    end process;
    en_dcalc  <= en_dcalc_stage(0) or en_dcalc_start;


    pCtrlComb: process(state, init_key, round, start, dcalc_mode,
        compute_round_r, is_i_done)
    begin
        --! External
        key_ready       <= '0';
        init_key_done_s <= '0';
        done_s          <= '0';
        --! internal        
        clr_round       <= '0';
        en_key          <= '0';
        en_round        <= '0';
        en_rami         <= '0';
        en_ramj         <= '0';
        en_key_init     <= '0';
        en_dcalc_start  <= '0';
        en_state        <= '0';
        sel_final       <= '0';
        sel_dcalc       <= "000";
        sel_key_init    <= "00";
        done_ram_init   <= '0';
        almost_done_s   <= '0';

        case state is
            when S_IDLE =>
                clr_round <= '1';
                if (unsigned(dcalc_mode) = natural(3)) then
                    sel_dcalc <= "001";
                elsif (unsigned(dcalc_mode) = natural(4)) then
                    sel_dcalc <= "010";
                elsif (unsigned(dcalc_mode) = natural(5)
                       or unsigned(dcalc_mode) = natural(6))
                then
                    sel_dcalc <= "011";
                else
                    sel_dcalc <= "100";
                end if;

                if (init_key = '1') then
                    nstate <= S_LD_KEY;                    
                elsif (start = '1') then
                    nstate <= S_COMPUTE;
                    en_dcalc_start <= '1';
                    en_state       <= '1';
                end if;
                
            when S_LD_KEY =>
                key_ready <= '1';                
                if key_valid = '1' then
                    en_key       <= '1';
                    en_round     <= '1';
                    if (round = "1011") then
                        clr_round <= '1';
                        nstate    <= S_INIT_I;
                    end if;
                end if;

            when S_INIT_I =>
                nstate       <= S_SETUP_IJ;
                sel_key_init <= "10";
                en_key_init  <= '1';

            when S_SETUP_IJ =>
                --! Populate I and J RAM with (2^1..16)I/J
                en_key_init  <= '1';
                en_round     <= '1';
                if (is_i_done = '0') then
                    en_rami      <= '1';
                else
                    en_ramj      <= '1';
                end if;
                if (round = "1111") then
                    done_ram_init <= '1';
                    if (is_i_done = '0') then
                        sel_key_init    <= "11";
                    else
                        nstate          <= S_IDLE;
                        done_s          <= '1';
                        init_key_done_s <= '1';
                    end if;
                end if;
                if (round = "1110" and is_i_done = '1') then
                    almost_done_s <= '1';
                end if;

            when S_COMPUTE =>
                en_state <= '1';
                en_round <= '1';
                if (round = compute_round_r) then
                    nstate    <= S_IDLE;
                    sel_final <= is_final;
                    clr_round <= '1';
                    done_s    <= '1';
                end if;
                if (unsigned(round) = unsigned(compute_round_r)-natural(1)) then
                    almost_done_s <= '1';
                end if;
                if (round = "0011") then
                    en_dcalc_start <= is_alpha;
                end if;
        end case;
    end process;
end architecture structure;
